"""Extra methods for the ``ee.String`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.String, "geetools")
class StringAccessor:
    """Toolbox for the ``ee.String`` class."""

    def __init__(self, obj: ee.String):
        """Initialize the String class."""
        self._obj = obj

    def eq(self, other: str | ee.String) -> ee.Number:
        """Compare two strings and return a ``ee.Number``.

        Parameters:
            other: The string to compare with.

        Returns:
            ``1`` if the strings are equal, ``0`` otherwise.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                s = ee.String("foo").geetools.eq("foo")
                s.getInfo()
        """
        return self._obj.compareTo(ee.String(other)).Not()

    def format(self, template: dict | ee.Dictionary) -> ee.String:
        """Format a string with a dictionary.

        Replace the keys in the string using the values provided in the dictionary. Follow the same pattern: value format as Python string.format method.

        Parameters:
            template: A dictionary with the values to replace.

        Returns:
            The formatted string.

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                s = ee.String("{greeting} {name} !").geetools.format({"greeting": "Hello", "name": "bob"})
                s.getInfo()
        """
        template = ee.Dictionary(template)
        templateList = template.keys().zip(template.values())

        def replace_format(kv, s):
            kv = ee.List(kv)
            key, value = ee.String(kv.get(0)), ee.String(kv.get(1))
            pattern = ee.String("{").cat(key).cat(ee.String("}"))
            return ee.String(s).replace(pattern, value)

        return ee.String(templateList.iterate(replace_format, self._obj))



"""Extra tools for the ``ee.DateRange`` class."""
from __future__ import annotations

import ee

from .accessors import register_class_accessor


@register_class_accessor(ee.DateRange, "geetools")
class DateRangeAccessor:
    """Toolbox for the ``ee.DateRange`` class."""

    def __init__(self, obj: ee.DateRange):
        """Initialize the DateRange class."""
        self._obj = obj

    # -- date range operations -------------------------------------------------
    def split(self, interval: int | ee.Number, unit: str = "day") -> ee.List:
        """Convert a ``ee.DateRange`` to a list of ``ee.DateRange``.

        The DateRange will be split in multiple DateRanges of the specified interval and Unit.
        For example "1", "day". if the end date is not included the last dateRange length will be adapted.

        Parameters:
            interval: The interval to split the DateRange
            unit: The unit to split the DateRange. One of: ``second``, ``minute``, ``hour``, ``day``, ``month``, ``year``.

        Returns:
            The list of DateRanges

        Examples:
            .. code-block:: python

                import ee, geetools

                ee.Initialize()

                d = ee.DateRange('2020-01-01', '2020-01-31').geetools.split(1, 'day')
                d.getInfo()
        """
        self.check_unit(unit)
        interval = ee.Number(interval).toInt().multiply(self.unitMillis(unit))
        start, end = self._obj.start().millis(), self._obj.end().millis()

        timestampList = ee.List.sequence(start, end, interval)
        timestampList = timestampList.add(ee.Number(end).toFloat()).distinct()
        indexList = ee.List.sequence(0, timestampList.size().subtract(2))

        return indexList.map(
            lambda i: ee.DateRange(timestampList.get(i), timestampList.get(ee.Number(i).add(1)))
        )

    # -- utils -----------------------------------------------------------------
    @staticmethod
    def check_unit(unit: str) -> None:
        """Check if the unit is valid."""
        if unit not in (units := ["second", "minute", "hour", "day", "month", "year"]):
            raise ValueError(f"unit must be one of: {','.join(units)}")

    @staticmethod
    def unitMillis(unit: str) -> ee.Number:
        """Get the milliseconds of a unit."""
        millis = {
            "second": 1000,
            "minute": 1000 * 60,
            "hour": 1000 * 60 * 60,
            "day": 1000 * 60 * 60 * 24,
            "month": 1000 * 60 * 60 * 24 * 30,
            "year": 1000 * 60 * 60 * 24 * 365,
        }
        return ee.Number(millis[unit])



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup ----------------------------------------------------------------
import os
import sys
from datetime import datetime
from pathlib import Path

# add . to sys to import local extensions
sys.path.append(str(Path(".").resolve()))

# -- Project information -------------------------------------------------------
project = "geetools"
author = "Rodrigo E. Principe"
copyright = f"2017-{datetime.now().year}, {author}"
release = "1.8.0"

# -- General configuration -----------------------------------------------------
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.icon",
    "sphinx_design",
    "sphinx_copybutton",
    "autoapi.extension",
    "jupyter_sphinx",
    "myst_nb",
    "_extension.docstring",
    "_extension.api_admonition",
]
exclude_patterns = ["**.ipynb_checkpoints"]

# -- Options for HTML output ---------------------------------------------------
# Define the json_url for our version switcher.
json_url = "https://geetools.readthedocs.io/en/latest/_static/switcher.json"

# Define the version we use for matching in the version switcher.
version_match = os.environ.get("READTHEDOCS_VERSION")

# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# for local development and the latest dev build use the local file instead of the distant one.
if not version_match or version_match.isdigit() or version_match == "latest":
    version_match = "dev"
    json_url = "_static/switcher.json"
elif version_match == "stable":
    version_match = f"v{release}"

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_logo = "_static/logo.png"
html_favicon = "_static/logo.png"
html_theme_options = {
    "logo": {
        "text": project,
    },
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/gee-community/geetools",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "Pypi",
            "url": "https://pypi.org/project/geetools/",
            "icon": "fa-brands fa-python",
            "type": "fontawesome",
        },
        {
            "name": "Conda",
            "url": "https://anaconda.org/conda-forge/geetools",
            "icon": "fa-custom fa-conda",
            "type": "fontawesome",
        },
    ],
    "announcement": "https://raw.githubusercontent.com/gee-community/geetools/main/docs/_static/banner.html",
    "secondary_sidebar_items": [
        "page-toc.html",
        "edit-this-page.html",
    ],
    # remove the switcher for now as the version management is not satisfying
    # "switcher": {
    #    "json_url": json_url,
    #    "version_match": version_match,
    # },
    "show_toc_level": 2,
}
html_context = {
    "github_user": "gee-community",
    "github_repo": "geetools",
    "github_version": "main",
    "doc_path": "docs",
}
html_css_files = ["custom.css"]
html_js_files = ["custom-icon.js"]

# -- Options for autosummary/autodoc output ------------------------------------
autodoc_typehints = "description"
autoapi_dirs = ["../geetools"]
autoapi_python_class_content = "both"
autoapi_member_order = "groupwise"
autoapi_template_dir = "_templates"
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
]
autoapi_own_page_level = "method"
autoapi_keep_files = False

# -- Options for intersphinx output --------------------------------------------
intersphinx_mapping = {}

# -- options for the autolabel extension ---------------------------------------
autosectionlabel_prefix_document = True

# -- options for myst-nb ------------------------------------------------------
nb_execution_mode = "auto"



# coding=utf-8
"""legacy Module holding tools for ee.ImageCollections."""
import ee
from deprecated.sphinx import deprecated

import geetools  # noqa: F401


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.closestDate instead.")
def fillWithLast(collection, reverse=False, proxy=-999):
    """Fill each masked pixels with the last available not masked pixel."""
    return ee.ImageCollection(collection).geetools.closestDate()


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.reduceInterval instead.")
def mosaicSameDay(collection, qualityBand=""):
    """Return a collection where images from the same day are mosaicked."""
    reducer = "mosaic" if qualityBand == "" else "qualityMosaic"
    return ee.ImageCollection(collection).geetools.reduceInterval(reducer, "day", 1, qualityBand)


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.reduceInterval instead.")
def reduceEqualInterval(
    collection, interval=30, unit="day", reducer=None, start_date=None, end_date=None
):
    """Reduce an ImageCollection into a new one that has one image per reduced interval."""
    return ee.ImageCollection(collection).geetools.reduceInterval(reducer, unit, interval)


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.groupInterval instead.")
def makeEqualInterval(collection, interval=1, unit="month"):
    """Make a list of image collections filtered by the given interval."""
    return ee.ImageCollection(collection).geetools.groupInterval(unit, interval)


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.groupInterval instead..")
def makeDayIntervals(collection, interval=30, reverse=False, buffer="second"):
    """Make day intervals."""
    return ee.ImageCollection(collection).geetools.groupInterval("day", 1)


@deprecated(version="1.5.0", reason="Use ee.ImageCollection.geetools.reduceInterval instead.")
def reduceDayIntervals(collection, reducer, interval=30, reverse=False, buffer="second"):
    """Reduce Day Intervals."""
    return ee.ImageCollection(collection).geetools.reduceInterval(reducer, "day", 1)


@deprecated(version="1.4.0", reason="Removed from the lib as untested")
def linearFunctionBand(*args, **kwargs):
    """Apply a linear function over the bands across every image of the ImageCollection"""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.4.0", reason="Removed from the lib as untested")
def linearFunctionProperty(*args, **kwargs):
    """Apply a linear function over the properties across every image of the ImageCollection"""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.4.0", reason="Removed from the lib as untested.")
def linearInterpolation(collection, date_property="system:time_start"):
    """TODO missing docstring."""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.4.0", reason="Removed from the lib as untested.")
def gaussFunctionBand(*args, **kwargs):
    """Compute a Gauss function using a specified band over an ImageCollection, See: https://en.wikipedia.org/wiki/Gaussian_function."""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.4.0", reason="Removed from the lib as untested.")
def gaussFunctionProperty(*args, **kwargs):
    """Compute a Gauss function using a specified property over an ImageCollection."""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.4.0", reason="Removed from the lib as untested.")
def normalDistributionProperty(*args, **kwargs):
    """Compute a normal distribution using a specified property, over an ImageCollection"""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.4.0", reason="Removed from the lib as untested.")
def normalDistributionBand(collection, band, mean=None, std=None, name="normal_distribution"):
    """Compute a normal distribution using a specified band, over an ImageCollection."""
    raise NotImplementedError(
        "As it was vastly untested this method has been removed from the lib."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )


@deprecated(version="1.5.0", reason="The output format is unclear.")
def moving_average(collection, back=5, reducer=None, use_original=True):
    """Compute the moving average over a time series."""
    raise NotImplementedError(
        "This method has been deprecated as the output format is unclear."
        "If a real use case for this method can be provided, please open an issue and we'll reimplement it."
    )



"""Test the ``Image`` class."""
import io
import zipfile
from io import BytesIO
from math import isclose
from pathlib import Path
from tempfile import TemporaryDirectory
from urllib.request import urlretrieve

import ee
import numpy as np
import pytest
from jsonschema import validate
from matplotlib import pyplot as plt

import geetools


class TestAddDate:
    """Test the ``addDate`` method."""

    def test_add_date(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.addDate()
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})

    def test_add_date_format(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.addDate("yyyyMMdd")
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})


class TestAddSuffix:
    """Test the ``addSuffix`` method."""

    def test_add_suffix_to_all(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.addSuffix("_suffix")
        data_regression.check(image.bandNames().getInfo())

    def test_add_suffix_to_selected(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.addSuffix("_suffix", bands=["B1", "B2"])
        data_regression.check(image.bandNames().getInfo())


class TestAddPrefix:
    """Test the ``addPrefix`` method."""

    def test_add_prefix_to_all(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.addPrefix("prefix_")
        data_regression.check(image.bandNames().getInfo())

    def test_add_prefix_to_selected(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.addPrefix("prefix_", bands=["B1", "B2"])
        data_regression.check(image.bandNames().getInfo())


class TestGetValues:
    """Test the ``getValues`` method."""

    def test_get_values(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        values = s2_sr_vatican_2020.geetools.getValues(vatican_buffer.centroid())
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})

    def test_get_values_with_scale(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        values = s2_sr_vatican_2020.geetools.getValues(vatican_buffer.centroid(), scale=100)
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})


class TestMinScale:
    """Test the ``minScale`` method."""

    def test_min_scale(self, s2_sr_vatican_2020):
        scale = s2_sr_vatican_2020.geetools.minScale()
        assert scale.getInfo() == 10


class TestMerge:
    """Test the ``merge`` method."""

    def test_merge(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.merge([s2_sr_vatican_2020, s2_sr_vatican_2020])
        data_regression.check(image.bandNames().getInfo())


class TestRename:
    """Test the ``rename`` method."""

    def test_rename(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.rename({"B1": "newB1", "B2": "newB2"})
        data_regression.check(image.bandNames().getInfo())


class TestRemove:
    """Test the ``remove`` method."""

    def test_remove(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.remove(["B1", "B2"])
        data_regression.check(image.bandNames().getInfo())


class TestToGrid:
    """Test the ``toGrid`` method."""

    def test_to_grid(self, s2_sr_vatican_2020, vatican_buffer, ndarrays_regression):
        grid = s2_sr_vatican_2020.geetools.toGrid(1, "B2", vatican_buffer)
        grid = [f["geometry"]["coordinates"] for f in grid.getInfo()["features"]]
        grid = {f"geometry_{i}": np.array(c) for i, c in enumerate(grid)}
        ndarrays_regression.check(grid)


class TestClipOnCollection:
    """Test the ``clipOnCollection`` method."""

    def test_clip_on_collection(self, s2_sr_vatican_2020, fc_instance, data_regression):
        clipped = s2_sr_vatican_2020.geetools.clipOnCollection(fc_instance)
        name = "test_clip_on_collection"
        data_regression.check(clipped.first().bandNames().getInfo(), f"{name}_bands")
        data_regression.check(clipped.first().propertyNames().getInfo(), f"{name}_property")

    def test_clip_on_collection_without_properties(
        self, s2_sr_vatican_2020, fc_instance, data_regression
    ):
        clipped = s2_sr_vatican_2020.geetools.clipOnCollection(fc_instance, 0)
        name = "test_clip_on_collection_without_properties"
        data_regression.check(clipped.first().bandNames().getInfo(), f"{name}_bands")
        data_regression.check(clipped.first().propertyNames().getInfo(), f"{name}_property")

    @pytest.fixture
    def fc_instance(self):
        """Return 2 little buffers in vaticanc city as a featurecollection."""
        return ee.FeatureCollection(
            [
                ee.Feature(ee.Geometry.Point([12.4534, 41.9029]).buffer(50), {"Id": 1}),
                ee.Feature(ee.Geometry.Point([12.4534, 41.9029]).buffer(100), {"Id": 2}),
            ]
        )


class TestBufferMask:
    """Test the ``bufferMask`` method."""

    @pytest.mark.xfail
    def test_buffer_mask(self):
        """I don't know what to test here."""
        assert False


class TestFull:
    """Test the ``full`` method."""

    def test_full(self, vatican_buffer, num_regression):
        image = ee.Image.geetools.full()
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_full_with_value(self, vatican_buffer, num_regression):
        image = ee.Image.geetools.full([1])
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_full_with_name(self, vatican_buffer, num_regression):
        image = ee.Image.geetools.full([1], ["toto"])
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_full_with_lists(self, vatican_buffer, num_regression):
        image = ee.Image.geetools.full([1, 2, 3], ["toto", "titi", "tata"])
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check(values.getInfo())


class TestFullLike:
    """Test the ``fullLike`` method."""

    def test_full_like(self, vatican_buffer, s2_sr_vatican_2020, num_regression):
        image = s2_sr_vatican_2020.set({"props": "toto"})
        image = image.geetools.fullLike(0)
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        assert "props" not in image.propertyNames().getInfo()
        num_regression.check(values.getInfo())

    def test_full_like_with_properties(self, s2_sr_vatican_2020):
        image = s2_sr_vatican_2020.set({"props": "toto"})
        image = image.geetools.fullLike(0, copyProperties=1)
        assert "props" in image.propertyNames().getInfo()

    def test_full_like_with_mask(self, s2_sr_vatican_2020, num_regression):
        image = s2_sr_vatican_2020.geetools.fullLike(0, keepMask=1)
        values = image.geetools.getValues(ee.Geometry.Point(0, 0))
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})


class TestReduceBands:
    """Test the ``reduceBands`` method."""

    def test_reduce_bands(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.reduceBands("sum")
        values = image.select("sum").reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_reduce_bands_with_bands(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.reduceBands("sum", ["B1", "B2"])
        values = image.select("sum").reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_reduce_bands_with_name(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.reduceBands("sum", name="toto")
        values = image.select("toto").reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())


class TestNegativeClip:
    """Test the ``negativeClip`` method."""

    def test_negative_clip(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.negativeClip(vatican_buffer)
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})


class testFormat:
    """Test the ``toString`` method."""

    def test_to_string(self, s2_sr_vatican_2020):
        string = s2_sr_vatican_2020.geetools.toString("date: {system_date}")
        assert string.getInfo() == "date: 2020-01-01"


class TestPrefixSuffix:
    """Test the ``prefix`` and ``suffix`` methods."""

    def test_prefix(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.addPrefix("prefix_")
        data_regression.check(image.bandNames().getInfo())

    def test_suffix(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.addSuffix("_suffix")
        data_regression.check(image.bandNames().getInfo())


class TestGauss:
    """Test the ``gauss`` method."""

    def test_gauss(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.gauss()
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_gauss_with_band(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.gauss("B2")
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())


class TestDoyToDate:
    """Test the ``doyToDate`` method."""

    def test_doy_to_date(self, doy_image, vatican_buffer, num_regression):
        image = doy_image.geetools.doyToDate(2023)
        values = image.reduceRegion(ee.Reducer.min(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_doy_to_date_with_format(self, doy_image, vatican_buffer, num_regression):
        image = doy_image.geetools.doyToDate(2023, dateFormat="yyyy.DDD")
        values = image.reduceRegion(ee.Reducer.min(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_doy_to_date_with_band(self, doy_image, vatican_buffer, num_regression):
        image = doy_image.geetools.doyToDate(2023, band="doy2")
        values = image.reduceRegion(ee.Reducer.min(), vatican_buffer, 10)
        num_regression.check(values.getInfo())


class TestRepeat:
    """Test the ``repeat`` method."""

    def test_repeat(self, image_instance):
        image = image_instance.geetools.repeat("B1", 2)
        assert image.bandNames().getInfo() == ["B1", "B2", "B3", "B1_1", "B1_2"]

    @pytest.fixture
    def image_instance(self):
        """Return an Image instance."""
        src = "COPERNICUS/S2_SR_HARMONIZED/20200101T100319_20200101T100321_T32TQM"
        return ee.Image(src).select(["B1", "B2", "B3"])


class TestmatchHistogram:
    """Test the ``histogramMatch`` method."""

    def test_histogram_match(self, image_source, image_target, vatican_buffer, num_regression):
        bands = {"R": "R", "G": "G", "B": "B"}
        image = image_source.geetools.matchHistogram(image_target, bands)
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    @pytest.fixture
    def dates(self):
        """The dates of my imagery."""
        return "2023-06-01", "2023-06-30"

    @pytest.fixture
    def image_source(self, vatican_buffer, dates):
        """image from the S2 copernicus program over vatican city."""
        return (
            ee.ImageCollection("COPERNICUS/S2_SR_HARMONIZED")
            .filterBounds(vatican_buffer)
            .filterDate(*dates)
            .first()
            .select("B4", "B3", "B2")
            .rename("R", "G", "B")
        )

    @pytest.fixture
    def image_target(self, vatican_buffer, dates):
        """image from the L8 Landsat program over vatican city."""
        return (
            ee.ImageCollection("LANDSAT/LC08/C02/T1_L2")
            .filterBounds(vatican_buffer)
            .filterDate(*dates)
            .first()
            .select("SR_B4", "SR_B3", "SR_B2")
            .rename("R", "G", "B")
        )


class TestRemoveZeros:
    """Test the ``removeZeros`` method."""

    def test_remove_zeros(self, image_instance, vatican_buffer):
        image = image_instance.geetools.removeZeros()
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        assert values.getInfo() == {"array": [1, 2]}

    @pytest.fixture
    def image_instance(self):
        """A random image instance with array data containing zeros."""
        return ee.Image([0, 1, 2]).toArray()


class TestInterpolateBands:
    """Test the ``interpolateBands`` method."""

    def test_interpolate_bands(self, image_instance, vatican_buffer, num_regression):
        image = image_instance.geetools.interpolateBands([0, 3000], [0, 30])
        values = image.reduceRegion(ee.Reducer.first(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    @pytest.fixture
    def image_instance(self, vatican_buffer):
        """A sentinel 2 single image centered on the vatican."""
        src = "COPERNICUS/S2_SR_HARMONIZED"
        return ee.ImageCollection(src).filterBounds(vatican_buffer).first().select(["B4", "B2"])


class TestIsletMask:
    """Test the ``isletMask`` method."""

    def test_islet_mask(self, image_instance, tmp_path, image_regression):
        image = image_instance.geetools.isletMask(20)
        file = self.get_image(image, tmp_path / "test.tif")
        image_regression.check(file.read_bytes())

    def get_image(self, image, dst):

        link = image.getDownloadURL(
            {
                "name": "test",
                "region": ee.Geometry.Point([12.4534, 41.9033]).buffer(1000),
                "filePerBand": False,
                "scale": 10,
            }
        )

        with TemporaryDirectory() as dir:
            tmp = Path(dir) / "tmp.zip"
            urlretrieve(link, tmp)
            with zipfile.ZipFile(tmp, "r") as zip_:
                dst.write_bytes(zip_.read(zip_.namelist()[0]))

        return dst

    @pytest.fixture
    def image_instance(self):
        """An image on top of the buffer."""
        buffer = ee.Geometry.Point([12.4534, 41.9033]).buffer(1000)
        return (
            ee.ImageCollection("COPERNICUS/S2_SR_HARMONIZED")
            .filterBounds(buffer)
            .filterDate("2023-01-01", "2023-01-31")
            .first()
            .select("B4", "B3", "B2")
        )


class TestIndicexList:
    """Test the ``index_list`` method."""

    def test_indices(self):
        indices = ee.Image.geetools.index_list()
        assert "NDVI" in indices.keys()
        assert len(indices) == 228


class TestSpectralIndices:
    """Test the ``spectralIndices`` method."""

    def test_default_spectral_indices(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.spectralIndices("all")
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check({k: np.nan if v is None else v for k, v in values.getInfo().items()})


class TestMaskClouds:
    """Test the ``maskClouds`` method."""

    def test_mask_S2_clouds(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        image = s2_sr_vatican_2020.geetools.maskClouds()
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        values = {k: np.nan if v is None else v for k, v in values.getInfo().items()}
        num_regression.check(values)


class TestGetscaleParams:
    """Test the ``getScaleParams`` method."""

    def test_get_scale_params(self, s2_sr_vatican_2020, data_regression):
        params = s2_sr_vatican_2020.geetools.getScaleParams()
        data_regression.check(params)


class TestGetOffsetParams:
    """Test the ``getOffsetParams`` method."""

    def get_offset_params(self, s2_sr_vatican_2020, data_regression):
        params = s2_sr_vatican_2020.geetools.getOffsetParams()
        data_regression.check(params)


class TestScaleAndOffset:
    """Test the ``scaleAndOffset`` method."""

    def test_scale_and_offset(self, vatican_buffer, s2_sr_vatican_2020, num_regression):
        image = s2_sr_vatican_2020.geetools.scaleAndOffset()
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())


class TestPreprocess:
    """Test the ``preprocess`` method."""

    def test_preprocess(self, vatican_buffer, s2_sr_vatican_2020, num_regression):
        image = s2_sr_vatican_2020.geetools.preprocess()
        values = image.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())


class TestGetSTAC:
    """Test the ``getSTAC`` method."""

    def test_get_stac_schema(self, s2_sr_vatican_2020, stac_schema):
        stac = s2_sr_vatican_2020.geetools.getSTAC()
        validate(stac, stac_schema)

    def test_get_stac(self, s2_sr_vatican_2020):
        stac = s2_sr_vatican_2020.geetools.getSTAC()
        assert stac["id"] == "COPERNICUS/S2_SR_HARMONIZED"


class TestGetDOI:
    """Test the ``getDOI`` method."""

    def get_doi(self, s2_sr_vatican_2020, data_regression):
        doi = s2_sr_vatican_2020.geetools.getDOI()
        data_regression.check(doi)


class TestGetCitation:
    """Test the ``getCitation`` method."""

    def get_citation(self, s2_sr_vatican_2020, data_regression):
        citation = s2_sr_vatican_2020.first().geetools.getCitation()
        data_regression.check(citation)


class TestPanSharpen:
    """Test the panSharpen method."""

    @pytest.mark.xfail(
        reason="This test is failing because the panSharpen method is not implemented for this platform."
    )
    def test_pan_sharpen(self, l8_sr_vatican_2020, num_regression):
        sharp = l8_sr_vatican_2020.geetools.panSharpen(
            method="HPFA", qa=["MSE", "RMSE"], maxPixels=1e13
        )
        centroid = sharp.geometry().centroid().buffer(100)
        values = sharp.reduceRegion(ee.Reducer.mean(), centroid, 1)
        num_regression.check(values.getInfo())


class TestTasseledCap:
    """Test the tasseledCap method."""

    @pytest.mark.xfail(
        reason="This test is failing because the tasseledCap method is not implemented for this platform."
    )
    def test_tasseled_cap(self, l8_sr_vatican_2020, num_regression):
        img = l8_sr_vatican_2020.geetools.tasseledCap()
        centroid = img.geometry().centroid().buffer(100)
        values = img.reduceRegion(ee.Reducer.mean(), centroid, 1)
        num_regression.check(values.getInfo())


class TestRemoveProperties:
    """Test the removeProperties method."""

    def test_remove_properties(self, s2_sr_vatican_2020, data_regression):
        image = s2_sr_vatican_2020.geetools.removeProperties(["system:time_start"])
        data_regression.check(image.propertyNames().getInfo())


class TestDistanceToMask:
    """Test the distanceToMask method."""

    def test_distance_to_mask(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        centerBuffer = vatican_buffer.centroid().buffer(100)
        BufferMask = ee.Image.constant(1).clip(centerBuffer)
        mask = ee.Image.constant(0).where(BufferMask, 1)
        distance = s2_sr_vatican_2020.geetools.distanceToMask(mask)
        values = distance.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_deprecated_distance_to_mask(self, s2_sr_vatican_2020, vatican_buffer, num_regression):
        centerBuffer = vatican_buffer.centroid().buffer(100)
        BufferMask = ee.Image.constant(1).clip(centerBuffer)
        mask = ee.Image.constant(0).where(BufferMask, 1)
        with pytest.deprecated_call():
            distance = geetools.algorithms.distanceToMask(s2_sr_vatican_2020, mask)
            values = distance.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
            num_regression.check(values.getInfo())


class TestDistance:
    """Test the ``distance`` method."""

    def test_distance(self, vatican_buffer, num_regression):
        # 2 images from june in vatican
        distance = self.image.geetools.distance(self.other)
        values = distance.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
        num_regression.check(values.getInfo())

    def test_deprecated_euclidian_distance(self, vatican_buffer, num_regression):
        # 2 images from june in vatican
        with pytest.deprecated_call():
            distance = geetools.algorithms.euclideanDistance(self.image, self.other)
            values = distance.reduceRegion(ee.Reducer.mean(), vatican_buffer, 10)
            num_regression.check(values.getInfo())

    @property
    def image(self):
        """Return an image from june in vatican."""
        image_id = "COPERNICUS/S2_SR_HARMONIZED/20210604T100029_20210604T100027_T32TQM"
        return ee.Image(image_id).select(["B4", "B3", "B2"])

    @property
    def other(self):
        """Return another image from june in vatican."""
        other_id = "COPERNICUS/S2_SR_HARMONIZED/20210604T100029_20210604T100027_T33TTG"
        return ee.Image(other_id).select(["B4", "B3", "B2"])


class TestMaskCover:
    """Test the ``maskCoverRegion`` method."""

    def test_mask_cover_region(self):
        aoi = ee.Geometry.Point([12.210900891755129, 41.928551351175386]).buffer(2200)
        ratio = self.image.geetools.maskCoverRegion(aoi, scale=10)
        assert isclose(ratio.getInfo(), 9.99, abs_tol=0.01)

    def test_mask_cover_region_zero(self):
        aoi = ee.Geometry.Point([11.880190936531116, 42.0159494554553]).buffer(1000)
        ratio = self.image.geetools.maskCoverRegion(aoi, scale=10)
        assert isclose(ratio.getInfo(), 0)

    def test_mask_cover_regions(self):
        geom = ee.Geometry.Point([12.210900891755129, 41.928551351175386]).buffer(2200)
        aoi = ee.FeatureCollection([ee.Feature(geom, {"test_property": 1})])
        result = self.image.geetools.maskCoverRegions(aoi, scale=10)
        feat = ee.Feature(result.first())
        ratio = feat.getInfo()["properties"]["mask_cover"]
        # ratio = ee.Number(feat.get('mask_cover'))
        # the last line should work, but it doesn't, I don't know why
        assert isclose(ratio, 9.99, abs_tol=0.01)

    def test_mask_cover_regions_zero(self):
        geom = ee.Geometry.Point([11.880190936531116, 42.0159494554553]).buffer(1000)
        aoi = ee.FeatureCollection([ee.Feature(geom, {"test_property": 1})])
        result = self.image.geetools.maskCoverRegions(aoi, scale=10)
        feat = ee.Feature(result.first())
        ratio = feat.getInfo()["properties"]["mask_cover"]
        # ratio = ee.Number(feat.get('mask_cover'))
        # the last line should work, but it doesn't, I don't know why
        assert isclose(ratio, 0)

    def test_deprecated_mask_cover(self):
        with pytest.deprecated_call():
            image = geetools.algorithms.maskCover(self.image)
            assert isclose(image.get("mask_cover").getInfo(), 18.04, rel_tol=0.01)

    @property
    def image(self):
        image_id = "COPERNICUS/S2_SR_HARMONIZED/20180401T100019_20180401T100022_T32TQM"
        image = ee.Image(image_id)
        qa = image.select("QA60")
        cloudBitMask, cirrusBitMask = 1 << 10, 1 << 11
        mask = qa.bitwiseAnd(cloudBitMask).eq(0).And(qa.bitwiseAnd(cirrusBitMask).eq(0))
        image = image.updateMask(mask)
        return image.select(["B4", "B3", "B2"])


class TestPlot:
    """Test the ``plot`` method."""

    def test_plot(self, s2_sr_vatican_2020, vatican, image_regression):
        fig, ax = plt.subplots()
        s2_sr_vatican_2020.geetools.plot(["B4", "B3", "B2"], vatican.geometry(), ax)

        with BytesIO() as image_byte:
            fig.savefig(image_byte, format="png")
            image_byte.seek(0)
            image_regression.check(image_byte.getvalue())

    def test_plot_one_band(self, s2_sr_vatican_2020, vatican, image_regression):
        fig, ax = plt.subplots()
        ndvi = s2_sr_vatican_2020.geetools.spectralIndices("NDVI")
        ndvi.geetools.plot(["NDVI"], vatican.geometry(), ax)

        with BytesIO() as image_byte:
            fig.savefig(image_byte, format="png")
            image_byte.seek(0)
            image_regression.check(image_byte.getvalue())

    def test_plot_one_band_cmap(self, s2_sr_vatican_2020, vatican, image_regression):
        fig, ax = plt.subplots()
        ndvi = s2_sr_vatican_2020.geetools.spectralIndices("NDVI")
        ndvi.geetools.plot(["NDVI"], vatican.geometry(), ax, cmap="RdYlGn")

        with BytesIO() as image_byte:
            fig.savefig(image_byte, format="png")
            image_byte.seek(0)
            image_regression.check(image_byte.getvalue())

    def test_plot_with_fc(self, s2_sr_vatican_2020, vatican, image_regression):
        fig, ax = plt.subplots()
        s2_sr_vatican_2020.geetools.plot(["B4", "B3", "B2"], vatican.geometry(), ax, fc=vatican)

        with BytesIO() as image_byte:
            fig.savefig(image_byte, format="png")
            image_byte.seek(0)
            image_regression.check(image_byte.getvalue())

    def test_plot_with_crs(self, s2_sr_vatican_2020, vatican, image_regression):
        fig, ax = plt.subplots()
        ndvi = s2_sr_vatican_2020.geetools.spectralIndices("NDVI")
        ndvi.geetools.plot(["NDVI"], vatican.geometry(), ax, crs="EPSG:3857", scale=10)

        with BytesIO() as image_byte:
            fig.savefig(image_byte, format="png")
            image_byte.seek(0)
            image_regression.check(image_byte.getvalue())


class TestFromList:
    """Test ``fromList`` method."""

    def test_from_list_unique(self):
        """Test using a list of unique band names."""
        sequence = ee.List([1, 2, 3])
        images = sequence.map(lambda i: ee.Image(ee.Number(i)).rename(ee.Number(i).int().format()))
        image = ee.Image.geetools.fromList(images)
        assert image.bandNames().getInfo() == ["1", "2", "3"]

    def test_from_list_multiband(self):
        """Test using a list of multiband images."""
        images = ee.List(
            [
                ee.Image([1, 2, 3]).rename(["1", "2", "3"]),
                ee.Image([4, 5]).rename(["4", "5"]),
            ]
        )
        image = ee.Image.geetools.fromList(images)
        assert image.bandNames().getInfo() == ["1", "2", "3", "4", "5"]


class TestPlotByRegions:
    """Test the ``plot_by_regions`` method."""

    def test_plot_by_regions_bar(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        self.image.geetools.plot_by_regions(
            type = "bar",
            regions = self.ecoregions,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ["01_tmean", "02_tmean", "03_tmean", "04_tmean", "05_tmean", "06_tmean", "07_tmean", "08_tmean", "09_tmean", "10_tmean", "11_tmean", "12_tmean"],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ['#604791', '#1d6b99', '#39a8a7', '#0f8755', '#76b349', '#f0af07', '#e37d05', '#cf513e', '#96356f', '#724173', '#9c4f97', '#696969'],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    def test_plot_by_regions_barh(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        self.image.geetools.plot_by_regions(
            type = "barh",
            regions = self.ecoregions,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ["01_tmean", "02_tmean", "03_tmean", "04_tmean", "05_tmean", "06_tmean", "07_tmean", "08_tmean", "09_tmean", "10_tmean", "11_tmean", "12_tmean"],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ['#604791', '#1d6b99', '#39a8a7', '#0f8755', '#76b349', '#f0af07', '#e37d05', '#cf513e', '#96356f', '#724173', '#9c4f97', '#696969'],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    def test_plot_by_regions_stacked(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        self.image.geetools.plot_by_regions(
            type = "stacked",
            regions = self.ecoregions,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ["01_tmean", "02_tmean", "03_tmean", "04_tmean", "05_tmean", "06_tmean", "07_tmean", "08_tmean", "09_tmean", "10_tmean", "11_tmean", "12_tmean"],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ['#604791', '#1d6b99', '#39a8a7', '#0f8755', '#76b349', '#f0af07', '#e37d05', '#cf513e', '#96356f', '#724173', '#9c4f97', '#696969'],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def ecoregions(self):
        return ee.FeatureCollection("projects/google/charts_feature_example").select(
            ["label", "value", "warm"]
        )

    @property
    def image(self):
        return ee.ImageCollection("OREGONSTATE/PRISM/Norm91m").toBands()


class TestPlotByBands:
    """Test the ``plot_by_bands`` method."""

    def test_plot_by_bands_bar(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        self.image.geetools.plot_by_bands(
            type = "bar",
            regions = self.ecoregions,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ['01_ppt', '02_ppt', '03_ppt', '04_ppt', '05_ppt', '06_ppt', '07_ppt', '08_ppt', '09_ppt', '10_ppt', '11_ppt', '12_ppt'],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ["#f0af07", "#0f8755", "#76b349"],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    def test_plot_by_bands_plot(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        self.image.geetools.plot_by_bands(
            type = "plot",
            regions = self.ecoregions,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ['01_ppt', '02_ppt', '03_ppt', '04_ppt', '05_ppt', '06_ppt', '07_ppt', '08_ppt', '09_ppt', '10_ppt', '11_ppt', '12_ppt'],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ["#f0af07", "#0f8755", "#76b349"],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    def test_plot_by_bands_area(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        self.image.geetools.plot_by_bands(
            type = "fill_between",
            regions = self.ecoregions,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ['01_ppt', '02_ppt', '03_ppt', '04_ppt', '05_ppt', '06_ppt', '07_ppt', '08_ppt', '09_ppt', '10_ppt', '11_ppt', '12_ppt'],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ["#f0af07", "#0f8755", "#76b349"],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    def test_plot_by_bands_pie(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        ecoregion = self.ecoregions.filter(ee.Filter.eq("label", "Forest"))
        self.image.geetools.plot_by_bands(
            type = "pie",
            regions = ecoregion,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ['01_ppt', '02_ppt', '03_ppt', '04_ppt', '05_ppt', '06_ppt', '07_ppt', '08_ppt', '09_ppt', '10_ppt', '11_ppt', '12_ppt'],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ['#604791', '#1d6b99', '#39a8a7', '#0f8755', '#76b349', '#f0af07', '#e37d05', '#cf513e', '#96356f', '#724173', '#9c4f97', '#696969'],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    def test_plot_by_bands_donut(self, image_regression):
        fig, ax = plt.subplots()
        # fmt: off
        ecoregion = self.ecoregions.filter(ee.Filter.eq("label", "Forest"))
        self.image.geetools.plot_by_bands(
            type = "donut",
            regions = ecoregion,
            reducer = "mean",
            scale = 500,
            regionId = "label",
            bands = ['01_ppt', '02_ppt', '03_ppt', '04_ppt', '05_ppt', '06_ppt', '07_ppt', '08_ppt', '09_ppt', '10_ppt', '11_ppt', '12_ppt'],
            labels = ["jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec"],
            colors = ['#604791', '#1d6b99', '#39a8a7', '#0f8755', '#76b349', '#f0af07', '#e37d05', '#cf513e', '#96356f', '#724173', '#9c4f97', '#696969'],
            ax = ax
        )
        # fmt: on
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def ecoregions(self):
        return ee.FeatureCollection("projects/google/charts_feature_example").select(
            ["label", "value", "warm"]
        )

    @property
    def image(self):
        return ee.ImageCollection("OREGONSTATE/PRISM/Norm91m").toBands()


class TestPlotHist:
    """Test the ``plot_hist`` method."""

    def test_plot_hist(self, image_regression):
        fig, ax = plt.subplots()
        self.image.geetools.plot_hist(
            bands=["sur_refl_b01", "sur_refl_b02", "sur_refl_b06"],
            labels=[["Red", "NIR", "SWIR"]],
            colors=["#cf513e", "#1d6b99", "#f0af07"],
            ax=ax,
            bins=100,
            scale=500,
            region=self.region,
        )
        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def image(self):
        return (
            ee.ImageCollection("MODIS/061/MOD09A1")
            .filter(ee.Filter.date("2018-06-01", "2018-09-01"))
            .select(["sur_refl_b01", "sur_refl_b02", "sur_refl_b06"])
            .mean()
        )

    @property
    def region(self):
        return ee.Geometry.Rectangle([-112.60, 40.60, -111.18, 41.22])



"""Pytest session configuration."""

import string

import ee
import pytest
import pytest_gee
import requests

S2_BAND_COMBO = ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8", "B8A", "B9", "B11", "B12", "SCL"]
"""Sentinel-2 band combination."""

L8_BAND_COMBO = ["B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8", "B9", "B10", "B11"]
"""Landsat-8 band combination."""


def pytest_configure() -> None:
    """Initialize earth engine according to the environment."""
    pytest_gee.init_ee_from_service_account()


@pytest.fixture(scope="session")
def gee_folder_structure():
    """Override the default test folder structure."""
    point = ee.Geometry.Point([0, 0])
    return {
        "folder::Folder": {
            "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            "subfolder::Folder": {
                "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            },
        },
        "rmdir_folder::Folder": {
            "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            "subfolder::Folder": {
                "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            },
        },
        "move_folder::Folder": {
            "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            "subfolder::Folder": {
                "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            },
        },
        "copy_folder::Folder": {
            "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            "subfolder::Folder": {
                "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
            },
        },
        "unlink_folder::Folder": {
            "image": ee.Image(1).clipToBoundsAndScale(point.buffer(100), scale=30),
        },
    }


# -- fixtures that will be used throughout the tests ---------------------------
@pytest.fixture
def amazonas() -> ee.FeatureCollection:
    """Return the Amazonas state from colombia."""
    level2 = ee.FeatureCollection("FAO/GAUL/2015/level2")
    colombia = level2.filter(ee.Filter.eq("ADM0_NAME", "Colombia"))
    return colombia.filter(ee.Filter.eq("ADM1_NAME", "Amazonas"))


@pytest.fixture
def s2_sr(amazonas) -> ee.ImageCollection:
    """Return a copernicus based collection.

    the 100 first images of the Sentinel-2 Surface Reflectance ImageCollection centered on the amazonas state of colombia and from 2021-01-01 to 2021-12-01.
    """
    return (
        ee.ImageCollection("COPERNICUS/S2_SR_HARMONIZED")
        .select(S2_BAND_COMBO)
        .filterBounds(amazonas)
        .filterDate("2021-01-01", "2021-12-01")
    )


@pytest.fixture
def vatican():
    """Return the vatican city."""
    level0 = ee.FeatureCollection("FAO/GAUL/2015/level0")
    return level0.filter(ee.Filter.eq("ADM0_NAME", "Holy See"))


@pytest.fixture
def vatican_buffer():
    """A 100 buffer around vatican city."""
    return ee.Geometry.Point([12.4534, 41.9033]).buffer(100)


@pytest.fixture
def s2_sr_vatican_2020():
    """A single image from 2020 on top of vatican city from S2 SR collection."""
    src = "COPERNICUS/S2_SR_HARMONIZED/20200101T100319_20200101T100321_T32TQM"
    return ee.Image(src).select(S2_BAND_COMBO)


@pytest.fixture
def l8_sr_vatican_2020():
    """A single image from 2020 on top of vatican city from L8 SR collection."""
    src = "LANDSAT/LC08/C02/T1/LC08_191031_20130711"
    return ee.Image(src).select(L8_BAND_COMBO)


@pytest.fixture
def l8_toa(amazonas) -> ee.ImageCollection:
    """Return a landsat based collection.

    the 100 first images of the landast 8 TOA ImageCollection centered on the amazonas state of colombia and from 2021-01-01 to 2021-12-01.
    """
    return (
        ee.ImageCollection("LANDSAT/LC08/C02/T1_RT_TOA")
        .select(L8_BAND_COMBO)
        .filterBounds(amazonas)
        .filterDate("2021-01-01", "2021-12-01")
    )


@pytest.fixture
def l8_sr(amazonas):
    """Return a landsat based collection.

    the 100 first images of the landast 8 SR ImageCollection centered on the amazonas state of colombia and from 2021-01-01 to 2021-12-01.
    """
    return (
        ee.ImageCollection("LANDSAT/LC08/C02/T1_L2")
        .filterBounds(amazonas)
        .filterDate("2021-01-01", "2021-12-01")
    )


@pytest.fixture
def l8_sr_raw():
    """Return a defined image collection."""
    return ee.ImageCollection("LANDSAT/LC08/C02/T1").select(L8_BAND_COMBO)


@pytest.fixture
def date_instance():
    """Return a defined date instance."""
    return ee.Date("2020-01-01")


@pytest.fixture
def daterange_instance():
    """Return a DateRange instance."""
    return ee.DateRange("2020-01-01", "2020-01-31")


@pytest.fixture
def letter_list():
    """Return a defined list instance."""
    return ee.List([*string.ascii_lowercase[:3]])


@pytest.fixture
def int_list():
    """Return a defined list instance."""
    return ee.List([*range(1, 3)])


@pytest.fixture
def mix_list():
    """Return a defined list instance."""
    return ee.List(["a", 1, ee.Image(1)])


@pytest.fixture
def number_instance():
    """Return a defined number instance."""
    return ee.Number(1234.56785678)


@pytest.fixture
def string_instance():
    """Return a defined string instance."""
    return ee.String("foo")


@pytest.fixture
def format_string_instance():
    """Return a defined string instance."""
    return ee.String("{greeting} {name} !")


@pytest.fixture
def geom_instance():
    """Set a geometryCollection instance."""
    point0 = ee.Geometry.Point([0, 0], proj="EPSG:4326")
    point1 = ee.Geometry.Point([0, 1], proj="EPSG:4326")
    poly0 = point0.buffer(1, proj="EPSG:4326")
    poly1 = point1.buffer(1, proj="EPSG:4326").bounds(proj="EPSG:4326")
    line = ee.Geometry.LineString([point1, point0], proj="EPSG:4326")
    multiPoly = ee.Geometry.MultiPolygon([poly0, poly1], proj="EPSG:4326")
    return ee.Algorithms.GeometryConstructors.MultiGeometry(
        [multiPoly, poly0, poly1, point0, line],
        crs="EPSG:4326",
        geodesic=True,
        maxError=1,
    )


@pytest.fixture
def fc_instance():
    """Return a fc collection containing 1 single geometryCollection."""
    point0 = ee.Geometry.Point([0, 0], proj="EPSG:4326")
    point1 = ee.Geometry.Point([0, 1], proj="EPSG:4326")
    poly0 = point0.buffer(1, proj="EPSG:4326")
    poly1 = point1.buffer(1, proj="EPSG:4326").bounds(proj="EPSG:4326")
    line = ee.Geometry.LineString([point1, point0], proj="EPSG:4326")
    multiPoly = ee.Geometry.MultiPolygon([poly0, poly1], proj="EPSG:4326")
    geometryCollection = ee.Algorithms.GeometryConstructors.MultiGeometry(
        [multiPoly, poly0, poly1, point0, line],
        crs="EPSG:4326",
        geodesic=True,
        maxError=1,
    )
    return ee.FeatureCollection([geometryCollection])


@pytest.fixture
def gaul_3_countries():
    """Return Italy switzerland and France."""
    fc = ee.FeatureCollection("FAO/GAUL/2015/level0")
    return fc.filter(ee.Filter.inList("ADM0_CODE", [122, 237, 85]))


@pytest.fixture
def doy_image():
    """Return an Image instance with 2 random doy bands."""
    doy = ee.Image.random(seed=0).multiply(365).toInt().rename("doy1")
    return doy.rename("doy1").addBands(doy.rename("doy2"))


@pytest.fixture
def multipoint_feature():
    """Return a ``Feature`` instance."""
    geoms = ee.Geometry.MultiPoint([[0, 0], [0, 1]])
    return ee.Feature(geoms).set({"foo": "bar", "bar": "foo"})


@pytest.fixture
def ecoregions():
    """Return the ecoregion collection."""
    return ee.FeatureCollection("projects/google/charts_feature_example")


@pytest.fixture
def climSamp():
    """Return the climate sample collection."""
    normClim = ee.ImageCollection("OREGONSTATE/PRISM/Norm81m").toBands()
    region = ee.Geometry.Rectangle(-123.41, 40.43, -116.38, 45.14)
    return normClim.sample(region, 5000)


@pytest.fixture(scope="session")
def stac_schema():
    """Return the STAC collection schema."""
    url = "https://raw.githubusercontent.com/radiantearth/stac-spec/v1.0.0/collection-spec/json-schema/collection.json"
    return requests.get(url).json()


@pytest.fixture(scope="session")
def jaxa_rainfall():
    """Return the JAXA rain collection."""
    return ee.ImageCollection("JAXA/GPM_L3/GSMaP/v6/operational")



"""Test the ImageCollection class."""
from __future__ import annotations

import io

import ee
import numpy as np
import pytest
from jsonschema import validate
from matplotlib import pyplot as plt

import geetools


def reduce(
    collection: ee.ImageCollection, geometry: ee.Geometry | None = None, reducer: str = "first"
) -> ee.Dictionary:
    """Compute the mean reduction on the first image of the imageCollection."""
    image = getattr(collection, reducer)()
    geometry = image.geometry() if geometry is None else geometry.geometry()
    geometry = geometry.centroid(1).buffer(100)
    return image.reduceRegion(ee.Reducer.mean(), geometry, 1)


class TestMaskClouds:
    """Test the ``maskClouds`` method."""

    @pytest.mark.xfail(
        reason="ee_extra is joining ImgeCollection which is not compatible with ee v1.x."
    )
    def test_mask_s2_sr(self, s2_sr, num_regression):
        masked = s2_sr.geetools.maskClouds(prob=75, buffer=300, cdi=-0.5)
        num_regression.check(reduce(masked).getInfo())


class TestClosest:
    """Test the ``closest`` method."""

    def test_closest_s2_sr(self, s2_sr, data_regression):
        closest = s2_sr.geetools.closest("2021-10-01")
        data_regression.check(closest.size().getInfo())


class TestSpectralIndices:
    """Test the ``spectralIndices`` method."""

    def test_spectral_indices(self, s2_sr, num_regression):
        indices = s2_sr.geetools.spectralIndices(["NDVI", "NDWI"])
        num_regression.check(reduce(indices).getInfo())


class TestGetScaleParams:
    """Test the ``getScaleParams`` method."""

    def test_get_scale_params(self, s2_sr, data_regression):
        scale_params = s2_sr.geetools.getScaleParams()
        data_regression.check(scale_params)


class TestGetOffsetParams:
    """Test the ``getOffsetParams`` method."""

    def test_get_offset_params(self, s2_sr, data_regression):
        offset_params = s2_sr.geetools.getOffsetParams()
        data_regression.check(offset_params)


class TestScaleAndOffset:
    """Test the ``scaleAndOffset`` method."""

    def test_scale_and_offset(self, s2_sr, num_regression):
        scaled = s2_sr.geetools.scaleAndOffset()
        num_regression.check(reduce(scaled).getInfo())


class TestPreprocess:
    """Test the ``preprocess`` method."""

    @pytest.mark.xfail(
        reason="ee_extra is joining ImgeCollection which is not compatible with ee v1.x."
    )
    def test_preprocess(self, s2_sr, num_regression):
        preprocessed = s2_sr.geetools.preprocess()
        values = {k: np.nan if v is None else v for k, v in reduce(preprocessed).getInfo().items()}
        num_regression.check(values)


class TestGetSTAC:
    """Test the ``getSTAC`` method."""

    def test_get_stac(self, s2_sr):
        stac = s2_sr.geetools.getSTAC()
        assert stac["id"] == "COPERNICUS/S2_SR_HARMONIZED"

    def test_get_stac_schema(self, s2_sr, stac_schema):
        stac = s2_sr.geetools.getSTAC()
        validate(stac, stac_schema)


class TestGetDOI:
    """Test the ``getDOI`` method."""

    def test_get_doi(self, s2_sr, data_regression):
        doi = s2_sr.geetools.getDOI()
        data_regression.check(doi)


class TestGetCitation:
    """Test the ``getCitation`` method."""

    def test_get_citation(self, s2_sr, data_regression):
        citation = s2_sr.geetools.getCitation()
        data_regression.check(citation)


class TestPanSharpen:
    """Test the ``panSharpen`` method."""

    @pytest.mark.xfail(reason="ee_extra does not accept C02 L08 collection yet.")
    def test_pan_sharpen(self, l8_toa, num_regression):
        sharpened = l8_toa.geetools.panSharpen()
        num_regression.check(reduce(sharpened).getInfo())


class TestTasseledCap:
    """Test the ``tasseledCap`` method."""

    def test_tasseled_cap(self, l8_sr, num_regression):
        tc = l8_sr.geetools.tasseledCap()
        num_regression.check(reduce(tc).getInfo())


class TestAppend:
    """Test the ``append`` method."""

    def test_append(self, s2_sr, data_regression):
        appended = s2_sr.geetools.append(s2_sr.first())
        data_regression.check(appended.size().getInfo())


class TestcollectionMask:
    """Test the ``collectionMask`` method."""

    def test_collection_mask(self, s2_sr, amazonas, num_regression):
        masked = s2_sr.geetools.collectionMask()
        num_regression.check(reduce(ee.ImageCollection([masked]), amazonas).getInfo())


class TestIloc:
    """Test the iloc class."""

    def test_iloc(self, s2_sr, num_regression):
        ic = ee.ImageCollection([s2_sr.geetools.iloc(0).subtract(s2_sr.first())])
        values = {k: np.nan if v is None else v for k, v in reduce(ic).getInfo().items()}
        num_regression.check(values)


class TestIntegral:
    """Test the ``integral`` method."""

    def test_integral(self, s2_sr, amazonas, num_regression):
        integral = s2_sr.limit(10).geetools.integral("B4").select("integral")
        ic = ee.ImageCollection([integral])
        values = {k: np.nan if v is None else v for k, v in reduce(ic, amazonas).getInfo().items()}
        num_regression.check(values)


class TestOutliers:
    """Test the ``outliers`` method."""

    def test_outliers(self, s2_sr, amazonas, num_regression):
        ic = s2_sr.limit(10).geetools.outliers()
        values = {k: np.nan if v is None else v for k, v in reduce(ic, amazonas).getInfo().items()}
        num_regression.check(values)

    def test_outliers_with_bands(self, s2_sr, amazonas, num_regression):
        ic = s2_sr.limit(10).geetools.outliers(bands=["B4", "B2"])
        values = {k: np.nan if v is None else v for k, v in reduce(ic, amazonas).getInfo().items()}
        num_regression.check(values)

    def test_outliers_with_sigma(self, s2_sr, amazonas, num_regression):
        ic = s2_sr.limit(10).geetools.outliers(sigma=3)
        values = {k: np.nan if v is None else v for k, v in reduce(ic, amazonas).getInfo().items()}
        num_regression.check(values)

    def test_outliers_with_drop(self, s2_sr, amazonas, num_regression):
        ic = s2_sr.limit(10).geetools.outliers(drop=True)
        values = {k: np.nan if v is None else v for k, v in reduce(ic, amazonas).getInfo().items()}
        num_regression.check(values)


class TestToXarray:
    """Test the ``toXarray`` method."""

    def test_to_xarray(self, s2_sr, data_regression):
        ds = s2_sr.geetools.to_xarray()

        # drop all the dtype as they are not consistently setup depending on the xarray version
        def drop_dtype(d=ds):
            for k, v in ds.items():
                if isinstance(v, dict):
                    drop_dtype(v)
                elif k == "dtype":
                    del ds[k]

        drop_dtype()

        # ds = ds.astype(np.float64)
        data_regression.check(ds.to_dict(data=False))


class TestValidPixel:
    """Test the ``validPixel`` method."""

    def test_validPixel(self, s2_sr, amazonas, num_regression):
        s2_sr = s2_sr.filterDate("2021-01-01", "2021-01-31")
        ic = ee.ImageCollection([s2_sr.geetools.validPixel("B1")])
        values = {k: np.nan if v is None else v for k, v in reduce(ic, amazonas).getInfo().items()}
        num_regression.check(values)


class TestContainsBandNames:
    """Test the ``containsBandNames`` method and derivated."""

    def test_contains_all(self, s2_sr):
        ic = s2_sr.select(["B2", "B3", "B4"])
        ic = ic.geetools.containsAllBands(["B2", "B3"])
        assert ic.size().getInfo() == 2449

    def test_contains_all_mismatch(self, s2_sr):
        ic = s2_sr.select(["B2", "B3", "B4"])
        ic = ic.geetools.containsAllBands(["B2", "B3", "B5"])
        assert ic.size().getInfo() == 0

    def test_contains_any(self, s2_sr):
        ic = s2_sr.select(["B2", "B3", "B4"])
        ic = ic.geetools.containsAnyBands(["B2", "B3", "B5"])
        assert ic.size().getInfo() == 2449

    def test_contains_any_mismatch(self, s2_sr):
        ic = s2_sr.select(["B2", "B3", "B4"])
        ic = ic.geetools.containsAnyBands(["B5", "B6"])
        assert ic.size().getInfo() == 0


class TestAggregateArray:
    """Test the ``aggregateArray`` method."""

    def test_aggregate_array(self, s2_sr, data_regression):
        # reduce the number of properties beforehand to avoid the test to fail
        keys = s2_sr.first().propertyNames()
        keys = keys.filter(ee.Filter.stringStartsWith("item", "system:")).remove("system:version")
        s2_sr_filtered = s2_sr.limit(3).map(
            lambda i: ee.Image().addBands(i).copyProperties(i, keys)
        )
        aggregated = s2_sr_filtered.geetools.aggregateArray()
        data_regression.check(aggregated.getInfo())

    def test_aggregate_array_with_properties(self, s2_sr, data_regression):
        aggregated = s2_sr.limit(10).geetools.aggregateArray(["system:time_start", "system:index"])
        data_regression.check(aggregated.getInfo())


class TestGroupInterval:
    """Test the ``groupInterval`` method."""

    def test_group_interval(self, jaxa_rainfall):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        grouped = ic.geetools.groupInterval()
        assert grouped.size().getInfo() == 3
        assert ee.ImageCollection(grouped.get(0)).size().getInfo() == 720

    def test_group_interval_with_interval(self, jaxa_rainfall):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        grouped = ic.geetools.groupInterval(duration=2)
        assert grouped.size().getInfo() == 2
        assert ee.ImageCollection(grouped.get(0)).size().getInfo() == 1440
        assert ee.ImageCollection(grouped.get(1)).size().getInfo() == 719

    def test_group_interval_with_interval_and_unit(self, jaxa_rainfall):
        # get 3 days worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-01-04")
        grouped = ic.geetools.groupInterval(duration=1, unit="day")
        assert grouped.size().getInfo() == 3
        assert ee.ImageCollection(grouped.get(0)).size().getInfo() == 24

    def test_deprecated_make_equal_interval(self, jaxa_rainfall):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        with pytest.deprecated_call():
            grouped = geetools.imagecollection.makeEqualInterval(ic)
            assert grouped.size().getInfo() == 3
            assert ee.ImageCollection(grouped.get(0)).size().getInfo() == 720

    def test_deprecated_make_day_intervals(self, jaxa_rainfall):
        # get 3 days worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-01-04")
        with pytest.deprecated_call():
            grouped = geetools.imagecollection.makeDayIntervals(ic)
            assert grouped.size().getInfo() == 3
            assert ee.ImageCollection(grouped.get(0)).size().getInfo() == 24


class TestReduceInterval:
    """Test the ``reduceInterval`` method."""

    def test_reduce_interval(self, jaxa_rainfall, amazonas, num_regression):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        reduced = ic.geetools.reduceInterval()
        values = {
            k: np.nan if v is None else v for k, v in reduce(reduced, amazonas).getInfo().items()
        }
        num_regression.check(values)

    def test_reduce_interval_with_reducer(self, jaxa_rainfall, amazonas, num_regression):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        reduced = ic.geetools.reduceInterval("max")
        values = reduce(reduced, amazonas).getInfo()
        values = {k: np.nan if v is None else v for k, v in values.items()}
        num_regression.check(values)

    def test_reduce_interval_with_non_existing_reducer_and_properties(self, jaxa_rainfall):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        with pytest.raises(AttributeError):
            ic.geetools.reduceInterval("toto")

    def test_deprecated_reduce_equal_interval(self, jaxa_rainfall, amazonas, num_regression):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-31")
        with pytest.deprecated_call():
            reduced = geetools.imagecollection.reduceEqualInterval(ic, reducer="mean")
            values = reduce(reduced, amazonas).getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)

    def test_deprecated_reduce_day_intervals(self, jaxa_rainfall, amazonas, num_regression):
        # get 3 days worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-01-04")
        with pytest.deprecated_call():
            reduced = geetools.imagecollection.reduceDayIntervals(ic, reducer="mean")
            values = reduce(reduced, amazonas).getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)

    def test_deprecated_composite_regular_intervals(self, jaxa_rainfall, amazonas, num_regression):
        # get 3 days worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-01-04")
        with pytest.deprecated_call():
            reduced = geetools.composite.compositeRegularIntervals(ic, unit="day")
            values = reduce(reduced, amazonas).getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)

    def test_deprecated_composite_by_month(self, jaxa_rainfall, amazonas, num_regression):
        # get 3 month worth of data and group it with default parameters
        ic = jaxa_rainfall.filterDate("2020-01-01", "2020-03-01")
        with pytest.deprecated_call():
            reduced = geetools.composite.compositeByMonth(ic)
            values = reduce(reduced, amazonas).getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)


class TestClosestDate:
    """Test the ``closestDate`` method."""

    def test_closest_date(self, s2_sr, amazonas, num_regression):
        # we need less images as the test will fail otherwise
        filled = s2_sr.filterDate("2021-01-01", "2021-01-15").geetools.closestDate()
        values = reduce(filled, amazonas, "mean").getInfo()
        values = {k: np.nan if v is None else v for k, v in values.items()}
        num_regression.check(values)

    def test_deprecated_fill_with_last(self, s2_sr, amazonas, num_regression):
        with pytest.deprecated_call():
            filled = geetools.imagecollection.fillWithLast(
                s2_sr.filterDate("2021-01-01", "2021-01-15")
            )
            values = reduce(filled, amazonas, "mean").getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)

    def test_deprecated_closest_date(self, s2_sr, amazonas, num_regression):
        with pytest.deprecated_call():
            filled = geetools.composite.closestDate(s2_sr.filterDate("2021-01-01", "2021-01-15"))
            values = reduce(filled, amazonas, "mean").getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)


class TestMedoid:
    """Test the ``medoid`` method."""

    def test_medoid(self, s2_sr, amazonas, num_regression):
        # we need less images as the test will fail otherwise
        medoid = s2_sr.filterDate("2021-01-01", "2021-01-05").geetools.medoid()
        values = reduce(ee.ImageCollection(medoid), amazonas).getInfo()
        values = {k: np.nan if v is None else v for k, v in values.items()}
        num_regression.check(values)

    def test_deprecated_medoid(self, s2_sr, amazonas, num_regression):
        with pytest.deprecated_call():
            # we need less images as the test will fail otherwise
            medoid = geetools.composite.medoid(s2_sr.filterDate("2021-01-01", "2021-01-05"))
            values = reduce(ee.ImageCollection(medoid), amazonas).getInfo()
            values = {k: np.nan if v is None else v for k, v in values.items()}
            num_regression.check(values)


class TestPlotDatesByBands:
    """Test the ``plot_dates_by_bands`` method."""

    def test_plot_dates_by_bands(self, image_regression):
        fig, ax = plt.subplots()
        self.collection.geetools.plot_dates_by_bands(
            region=self.region.geometry(),
            reducer="mean",
            scale=500,
            bands=["NDVI", "EVI"],
            ax=ax,
            dateProperty="system:time_start",
        )

        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def region(self):
        return (
            ee.FeatureCollection("projects/google/charts_feature_example")
            .select(["label", "value", "warm"])
            .filter(ee.Filter.eq("label", "Forest"))
        )

    @property
    def collection(self):
        return (
            ee.ImageCollection("MODIS/061/MOD13A1")
            .filter(ee.Filter.date("2010-01-01", "2020-01-01"))
            .select(["NDVI", "EVI"])
        )


class TestPlotDatesByRegions:
    """Test the ``plot_dates_by_regions`` method."""

    def test_plot_dates_by_regions(self, image_regression):
        fig, ax = plt.subplots()
        self.collection.geetools.plot_dates_by_regions(
            regions=self.regions,
            label="label",
            band="NDVI",
            reducer="mean",
            scale=500,
            ax=ax,
            dateProperty="system:time_start",
            colors=["#f0af07", "#0f8755", "#76b349"],
        )

        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def regions(self):
        return ee.FeatureCollection("projects/google/charts_feature_example").select(
            ["label", "value", "warm"]
        )

    @property
    def collection(self):
        return (
            ee.ImageCollection("MODIS/061/MOD13A1")
            .filter(ee.Filter.date("2010-01-01", "2020-01-01"))
            .select(["NDVI", "EVI"])
        )


class TestPlotDoyByBands:
    """Test the ``plot_doy_by_bands`` method."""

    def test_plot_doy_by_bands(self, image_regression):
        fig, ax = plt.subplots()
        self.collection.geetools.plot_doy_by_bands(
            region=self.region.geometry(),
            spatialReducer="mean",
            timeReducer="mean",
            scale=500,
            bands=["NDVI", "EVI"],
            ax=ax,
            dateProperty="system:time_start",
            colors=["#e37d05", "#1d6b99"],
        )

        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def region(self):
        return (
            ee.FeatureCollection("projects/google/charts_feature_example")
            .select(["label", "value", "warm"])
            .filter(ee.Filter.eq("label", "Grassland"))
        )

    @property
    def collection(self):
        return (
            ee.ImageCollection("MODIS/061/MOD13A1")
            .filter(ee.Filter.date("2010-01-01", "2020-01-01"))
            .select(["NDVI", "EVI"])
        )


class TestPlotDoyByRegions:
    """Test the ``plot_doy_by_regions`` method."""

    def test_plot_doy_by_regions(self, image_regression):
        fig, ax = plt.subplots()
        self.collection.geetools.plot_doy_by_regions(
            regions=self.regions,
            label="label",
            band="NDVI",
            spatialReducer="mean",
            timeReducer="mean",
            scale=500,
            ax=ax,
            dateProperty="system:time_start",
            colors=["#f0af07", "#0f8755", "#76b349"],
        )

        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def regions(self):
        return ee.FeatureCollection("projects/google/charts_feature_example").select(
            ["label", "value", "warm"]
        )

    @property
    def collection(self):
        return (
            ee.ImageCollection("MODIS/061/MOD13A1")
            .filter(ee.Filter.date("2010-01-01", "2020-01-01"))
            .select(["NDVI", "EVI"])
        )


class TestPlotDoyByYears:
    """Test the ``plot_doy_by_years`` method."""

    def test_plot_doy_by_years(self, image_regression):
        fig, ax = plt.subplots()
        self.collection.geetools.plot_doy_by_years(
            region=self.region.geometry(),
            band="NDVI",
            reducer="mean",
            scale=500,
            ax=ax,
            colors=["#39a8a7", "#9c4f97"],
        )

        with io.BytesIO() as buffer:
            fig.savefig(buffer)
            image_regression.check(buffer.getvalue())

    @property
    def region(self):
        return (
            ee.FeatureCollection("projects/google/charts_feature_example")
            .select(["label", "value", "warm"])
            .filter(ee.Filter.eq("label", "Grassland"))
        )

    @property
    def collection(self):
        return (
            ee.ImageCollection("MODIS/061/MOD13A1")
            .select(["NDVI", "EVI"])
            .filter(
                ee.Filter.Or(
                    ee.Filter.date("2012-01-01", "2012-12-31"),
                    ee.Filter.date("2019-01-01", "2019-12-31"),
                )
            )
        )



"""Test the ``Join`` class."""
import ee
import pytest


class TestByProperty:
    """Test the ``byProperty`` method."""

    def test_by_property(self, fc1, fc2, data_regression):
        joined = ee.Join.geetools.byProperty(fc1, fc2, "id")
        data_regression.check(joined.getInfo())

    def test_by_property_outer(self, fc1, fc2, data_regression):
        joined = ee.Join.geetools.byProperty(fc1, fc2, "id", outer=True)
        data_regression.check(joined.getInfo())

    @pytest.fixture
    def fc1(self):
        point = ee.Geometry.Point([0, 0])
        prop1 = {"id": 1, "prop_from_fc1": "I am from fc1"}
        return ee.FeatureCollection([ee.Feature(point, prop1)])

    @pytest.fixture
    def fc2(self):
        point = ee.Geometry.Point([0, 0])
        prop2 = {"id": 1, "prop_from_fc2": "I am from fc2"}
        return ee.FeatureCollection([ee.Feature(point, prop2)])



"""test the Number class methods."""


class TestTruncate:
    """Test the truncate method."""

    def test_truncate_with_default_decimals(self, number_instance):
        truncated_number = number_instance.geetools.truncate()
        assert truncated_number.getInfo() == 1234.56

    def test_truncate_with_custom_decimals(self, number_instance):
        truncated_number = number_instance.geetools.truncate(1)
        assert truncated_number.getInfo() == 1234.5

    def test_truncate_with_zero_decimals(self, number_instance):
        truncated_number = number_instance.geetools.truncate(0)
        assert truncated_number.getInfo() == 1234.0

    def test_truncate_with_large_decimals(self, number_instance):
        truncated_number = number_instance.geetools.truncate(5)
        assert truncated_number.getInfo() == 1234.56785



"""Test the ee_profiler module."""
import ee

import geetools  # noqa: F401


class TestProfiler:
    """Test the Profiler class."""

    def test_profiler(self):
        """Test the Profiler class."""

    with ee.geetools.Profiler() as p:
        ee.Number(3.14).add(0.00159).getInfo()
    assert [k for k in p.profile] == ["EECU-s", "PeakMem", "Count", "Description"]



"""Test the ``Geometry`` class."""


class TestKeepType:
    """Test the ``keepType`` method."""

    def test_keep_type(self, geom_instance, data_regression):
        geom = geom_instance.geetools.keepType("LineString")
        geojson = geom.getInfo()
        assert geojson["type"] == "MultiLineString"
        assert geom.coordinates().getInfo() == [[[0, 1], [0, 0]]]



"""Test the ComputedObject class methods."""

import ee
import pytest


class TestIsinstance:
    """Test the isInstance method."""

    def test_isinstance_with_string(self):
        assert ee.String("").isInstance(ee.String).getInfo() == 1

    def test_isinstance_with_integer(self):
        assert ee.Number(1).isInstance(ee.Integer).getInfo() == 1

    def test_isinstance_with_float(self):
        assert ee.Number(1.1).isInstance(ee.Float).getInfo() == 1

    def test_isinstance_with_image(self):
        assert ee.Image().isInstance(ee.Image).getInfo() == 1

    def test_isinstance_with_imagecollection(self):
        ic = ee.ImageCollection([ee.Image()])
        assert ic.isInstance(ee.ImageCollection).getInfo() == 1

    def test_isinstance_with_feature(self):
        assert ee.Feature(None).isInstance(ee.Feature).getInfo() == 1

    def test_isinstance_with_geometry(self):
        assert ee.Geometry.Point([0, 0]).isInstance(ee.Geometry).getInfo() == 1


class TestSave:
    """Test the ``save`` method."""

    def test_save(self, tmp_path):
        file = tmp_path / "test.gee"
        ee.Number(1.1).save(file)
        assert file.exists()


class TestOpen:
    """Test the ``open`` method."""

    def test_open(self, tmp_path):
        (object := ee.Number(1.1)).save((file := tmp_path / "test.gee"))
        opened = ee.Number.open(file)
        assert object.eq(opened).getInfo()

    def test_open_not_correct_suffix(self):
        with pytest.raises(ValueError):
            ee.Number.open("file.toto")



"""Test the String class methods."""
import ee


class TestEq:
    """Test the eq method."""

    def test_eq_with_same_string(self, string_instance):
        eq_number = string_instance.geetools.eq("foo")
        assert eq_number.getInfo() == 1

    def test_eq_with_different_string(self, string_instance):
        eq_number = string_instance.geetools.eq("bar")
        assert eq_number.getInfo() == 0


class TestFormat:
    """test the format method."""

    def test_format_with_dictionary(self, format_string_instance):
        params = {"greeting": "Hello", "name": "bob"}
        formatted_string = format_string_instance.geetools.format(params)
        assert formatted_string.getInfo() == "Hello bob !"

    def test_with_number(self, format_string_instance):
        params = {"greeting": "Hello", "name": ee.Number(1)}
        formatted_string = format_string_instance.geetools.format(params)
        assert formatted_string.getInfo() == "Hello 1 !"



"""Tools for the ``ee.Initialize`` function."""
from __future__ import annotations

import json
from pathlib import Path

import ee
from google.oauth2.credentials import Credentials

from .accessors import register_function_accessor

_project_id: str | None = None
"The project Id used by the current user."


@register_function_accessor(ee.Initialize, "geetools")
class InitializeAccessor:
    """Toolbox for the ``ee.Initialize`` function."""

    @staticmethod
    def from_user(name: str = "", credential_pathname: str = "", project: str = "") -> None:
        """Initialize Earthengine API using a specific user.

        Equivalent to the ``ee.initialize`` function but with a specific credential file stored in the machine by the ``ee.Authenticate.to_user`` function.

        Args:
            name: The name of the user as saved when created. use default if not set
            credential_pathname: The path to the folder where the credentials are stored. If not set, it uses the default path
            project: The project_id to use. If not set, it uses the default project_id of the saved credentials.

        Example:
            .. code-block:: python

                import ee
                import geetools

                ee.Initialize.from_user("test")

                # check that GEE is connected
                ee.Number(1).getInfo()
        """
        # gather global variable to be modified
        global _project_id

        # set the user profile information
        name = f"credentials{name}"
        credential_pathname = credential_pathname or ee.oauth.get_credentials_path()
        credential_folder = Path(credential_pathname).parent
        credential_path = credential_folder / name

        # check if the user exists
        if not credential_path.exists():
            msg = "Please register this user first by using geetools.User.create first"
            raise ee.EEException(msg)

        # Set the credential object and Init GEE API
        tokens = json.loads((credential_path / name).read_text())
        credentials = Credentials(
            None,
            refresh_token=tokens["refresh_token"],
            token_uri=ee.oauth.TOKEN_URI,
            client_id=tokens["client_id"],
            client_secret=tokens["client_secret"],
            scopes=ee.oauth.SCOPES,
        )
        ee.Initialize(credentials)

        # save the project_id in a dedicated global variable as it's not saved
        # from GEE side
        _project_id = project or tokens["project_id"]

    @staticmethod
    def project_id() -> str:
        """Get the project_id of the current account.

        Returns:
            The project_id of the connected profile

        Raises:
            RuntimeError: If the account is not initialized.

        Examples:
            .. code-block::

                import ee, geetools

                ee.Initialize.geetools.project_id()
        """
        if _project_id is None:
            raise RuntimeError("The GEE account is not initialized")
        return _project_id



"""Test the Filter class methods."""
import ee


class TestDateRange:
    """Test the dateRange method."""

    def test_dateRange_with_daterange(self, l8_sr_raw):
        filter = ee.Filter.geetools.dateRange(ee.DateRange("2018-01-01", "2019-01-01"))
        filtered_col = l8_sr_raw.filter(filter)
        assert filtered_col.size().getInfo() == 165030



"""Test the utils module."""

from geetools import utils


class TestFormatDescription:
    """Test the utils.format_description function."""

    def test_valid_description(self):
        """Test a valid description."""
        description = "This is a valid description 123.,:;_-"
        result = utils.format_description(description)
        assert result == "This_is_a_valid_description_123.,:;_-"

    def test_replacements(self):
        """Test replacements."""
        description = "Testing / replacements ?!{}()"
        result = utils.format_description(description)
        assert result == "Testing_-_replacements_..::::"

    def test_long_description(self):
        description = "A" * 150
        result = utils.format_description(description)
        assert len(result) == 100

    def test_unicode_characters(self):
        description = "Unicode characters like é, ä, and ñ should be changed"
        result = utils.format_description(description)
        assert result == "Unicode_characters_like_e,_a,_and_n_should_be_changed"


class TestFormatAssetID:
    """Test the utils.format_asset_id function."""

    def test_valid_description(self):
        """Test a valid description."""
        description = "This is a valid description 123.,:;_-"
        result = utils.format_asset_id(description)
        assert result == "This_is_a_valid_description_123_____-"

    def test_replacements(self):
        """Test replacements."""
        description = "Testing / replacements ?!{}()"
        result = utils.format_asset_id(description)
        assert result == "Testing_-_replacements_______"

    def test_long_description(self):
        description = "A" * 150
        result = utils.format_asset_id(description)
        assert len(result) == 150

    def test_unicode_characters(self):
        description = "Unicode characters like é, ä, and ñ should be changed"
        result = utils.format_asset_id(description)
        assert result == "Unicode_characters_like_e__a__and_n_should_be_changed"
